#ifndef FERI_ATTRACTOR
#define FERI_ATTRACTOR
// -----------------------------------------------
// ATTRACTOR ANALYSIS WITH GRAPH REPRESENTATION
// KRISZTIN-WALTHER MAP
//
// Main Program
// Ferenc Bartha
//
// using libraries:
// - CAPD
// - BOOST-GRAPH

// -----------------------------------------------

// GRAPHICS
//#define FERI_GRAPHICS

// GRAPHICS : DRAW EXTRA INFORMATION
//#define FERI_GRAPHICS_DRAW_EXTRA

// -----------------------------------------------

#include "main.hpp"
#include "typedefs.hpp"
#include "routines.hpp"
#include "krisztinwalther.hpp"

// -----------------------------------------------

// vectorfield
typedef VectorField<interval, FERI_DIMENSION> VectorFieldType;

// -----------------------------------------------

// trying to prove for a given parameter range
void simulateKW(interval param_m, interval param_alpha, double initialBoxSize, const std::string successFile, const std::string errorFile)
{
	boost::timer kwTime;
	VectorType V, enclosure, attDomain;
	GraphType GRAPH(0);
	double boxSize, runTime;
	int iter, max_iter;
	bool stop;



	// max number of iterations, 0 = unlimited
	max_iter = 100;



	VectorFieldType TheMap(param_m, param_alpha);



	#ifdef FERI_GRAPHICS
		openGraphics(500,500,WHITE,BLACK);
		double SCREENSIZE = 8;
		rootFrame.setWorldCoord(- SCREENSIZE / 2., - SCREENSIZE / 2., SCREENSIZE / 2., SCREENSIZE / 2.);

		TheMap.resetGraphics(SCREENSIZE);
	#endif



	// output
	#ifdef FERI_OUTPUT_PER_PARAM
		std::string output = outputBase + "_" + boost::lexical_cast<std::string>(param_m.leftBound()) + "_" + boost::lexical_cast<std::string>(param_m.rightBound()) + "_" + boost::lexical_cast<std::string>(param_alpha.leftBound()) + "_" + boost::lexical_cast<std::string>(param_alpha.rightBound()) + ".txt";

		std::ofstream ostream(output.c_str());
		ostream.precision(oPrecision);

		ostream << "Simulation for Krisztin-Walther Map, with parameters m = " << param_m << " and alpha = " << param_alpha << std::endl;
	#endif



	// we find the invariant set
	// we have finite convergence and stopping tolerance
	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "Constructing the invariant set... " << std::flush;
	#endif

	TheMap.constructInvariantSet(V);

	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "done" << std::endl;
		ostream << "Starting simulation on the invariant set X x Y: " << std::endl;
		ostream << "X = " << V(1) << std::endl;
		ostream << "Y = " << V(2) << std::endl;
	#endif



	// finding the local domain of attraction
	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "Constructing the local attraction domain... " << std::flush;
	#endif

	TheMap.findAttractionDomain(attDomain);

	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "done" << std::endl;
		ostream << "Local attraction domain is: " << attDomain << std::endl;
	#endif



	// building the initial graph
	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "Filling the initial graph with boxes... " << std::flush;
	#endif

	buildInitialBoxes(V, initialBoxSize, GRAPH);

	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "done" << std::endl;
	#endif



	// adding edges
	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "Adding edges to the graph... " << std::flush;
	#endif

	addInitialEdges(TheMap, GRAPH);

	#ifdef FERI_OUTPUT_PER_PARAM
		ostream << "done" << std::endl;
		reportSizeOfGraph(GRAPH, ostream);



		// evaluating enclosure for the invariant set
		evaluateEnclosure(boxSize, GRAPH, enclosure);
		ostream << "  ---- Enclosure is: " << enclosure << " with box size: " << boxSize << std::endl;
	#endif



	// MAIN LOOP STARTS
	iter = 0;
	do {
		++iter;

		#ifdef FERI_GRAPHICS
			TheMap.resetGraphics(SCREENSIZE);
		#endif



		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << " -------------- Iteration no. " << iter << std::endl;


			#ifdef FERI_SOURCESANDSINKS
				ostream << "Removing boxes that are sources or sinks... " << std::flush;
			#else
				ostream << "Keeping only SCCs... " << std::flush;
			#endif
		#endif



		#ifdef FERI_SOURCESANDSINKS
			removeSourcesAndSinks(GRAPH, 5);
		#else
			checkStronglyConnectedComponents(GRAPH);
		#endif



		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "done" << std::endl;
			reportSizeOfGraph(GRAPH, ostream);
		#endif



		// bisecting remaining boxes
		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "Bisecting boxes... " << std::flush;
		#endif

		bisectBoxes(GRAPH);

		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "done" << std::endl;
			reportSizeOfGraph(GRAPH, ostream);
		#endif



		// checking the mapping between new boxes
		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "Keeping only the good edges... " << std::flush;
		#endif

		keepOnlyGoodEdges(TheMap, GRAPH);

		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "done" << std::endl;
			reportSizeOfGraph(GRAPH, ostream);



			// evaluating enclosure for the invariant set
			if (evaluateEnclosure(boxSize, GRAPH, enclosure)) {
				ostream << "  ---- Enclosure is: " << enclosure << " with box size: " << boxSize << std::endl;
			}
		#endif



		#ifdef FERI_GRAPHICS

			if (evaluateEnclosure(boxSize, GRAPH, enclosure)) {

				SCREENSIZE = capd::max(capd::abs(enclosure(1)).rightBound(), capd::abs(enclosure(2)).rightBound());

				rootFrame.setWorldCoord(- SCREENSIZE / 2., - SCREENSIZE / 2., SCREENSIZE / 2., SCREENSIZE / 2.);

				plotBoxes(GRAPH, 1, 2);

				#ifdef FERI_GRAPHICS_DRAW_EXTRA
					TheMap.plotFixedPoint(SCREENSIZE);
				#endif
			}

		#endif



		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "Checking the attracted boxes... " << std::flush;
		#endif



		// checking if all the boxes are attracted
		stop = checkAttractionDomain(attDomain, GRAPH);



		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "done" << std::endl;
		#endif



	} while( not(stop) && ((iter < max_iter) || (max_iter == 0)) );
	// MAIN LOOP DONE



	runTime = kwTime.elapsed();



	// reached max number of iterations
	if ( not(stop) ) {

		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "Running Time : " << runTime << std::endl;
			ostream << "Could not validate Fixed Point" << std::endl;

			ostream.close();
		#endif

		std::ofstream ostream(errorFile.c_str(), std::ios::app);

		ostream.precision(oPrecision);

		if (evaluateEnclosure(boxSize, GRAPH, enclosure)) {
			ostream.width(0);
			ostream << "m = [";
			ostream.width(oPrecision + 2);
			ostream << param_m.leftBound();
			ostream.width(0);
			ostream << ",";
			ostream.width(oPrecision + 2);
			ostream << param_m.rightBound();
			ostream.width(0);
			ostream << "] a = [";
			ostream.width(oPrecision + 2);
			ostream << param_alpha.leftBound();
			ostream.width(0);
			ostream << ",";
			ostream.width(oPrecision + 2);
			ostream << param_alpha.rightBound();
			ostream.width(0);
			ostream << "]";
			ostream.width(oPrecision + 8);
			ostream << runTime;
			ostream.width(19);
			ostream << iter;
			ostream.width(28);
			ostream << boxSize;
			ostream.width(0);
			ostream << "      " << enclosure << std::endl;
		} else {
			ostream.width(0);
			ostream << "m = [";
			ostream.width(oPrecision + 2);
			ostream << param_m.leftBound();
			ostream.width(0);
			ostream << ",";
			ostream.width(oPrecision + 2);
			ostream << param_m.rightBound();
			ostream.width(0);
			ostream << "] a = [";
			ostream.width(oPrecision + 2);
			ostream << param_alpha.leftBound();
			ostream.width(0);
			ostream << ",";
			ostream.width(oPrecision + 2);
			ostream << param_alpha.rightBound();
			ostream.width(0);
			ostream << "]";
			ostream.width(oPrecision + 8);
			ostream << runTime;
			ostream.width(19);
			ostream << iter << std::endl;
		}

		ostream << std::flush;

		ostream.close();



	// fixed point is validated
	} else {

		#ifdef FERI_OUTPUT_PER_PARAM
			ostream << "Running Time : " << runTime << std::endl;
			ostream << "Attracting Fixed Point validated" << std::endl;

			ostream.close();
		#endif

		std::ofstream ostream(successFile.c_str(), std::ios::app);

		ostream.precision(oPrecision);

		if (evaluateEnclosure(boxSize, GRAPH, enclosure)) {
			ostream.width(0);
			ostream << "m = [";
			ostream.width(oPrecision + 2);
			ostream << param_m.leftBound();
			ostream.width(0);
			ostream << ",";
			ostream.width(oPrecision + 2);
			ostream << param_m.rightBound();
			ostream.width(0);
			ostream << "] a = [";
			ostream.width(oPrecision + 2);
			ostream << param_alpha.leftBound();
			ostream.width(0);
			ostream << ",";
			ostream.width(oPrecision + 2);
			ostream << param_alpha.rightBound();
			ostream.width(0);
			ostream << "]";
			ostream.width(oPrecision + 8);
			ostream << runTime;
			ostream.width(19);
			ostream << iter;
			ostream.width(28);
			ostream << boxSize;
			ostream.width(0);
			ostream << "      " << enclosure << std::endl;
		} else {
			ostream.width(0);
			ostream << "m = [";
			ostream.width(oPrecision + 2);
			ostream << param_m.leftBound();
			ostream.width(0);
			ostream << ",";
			ostream.width(oPrecision + 2);
			ostream << param_m.rightBound();
			ostream.width(0);
			ostream << "] a = [";
			ostream.width(oPrecision + 2);
			ostream << param_alpha.leftBound();
			ostream.width(0);
			ostream << ",";
			ostream.width(oPrecision + 2);
			ostream << param_alpha.rightBound();
			ostream.width(0);
			ostream << "]";
			ostream.width(oPrecision + 8);
			ostream << runTime;
			ostream.width(19);
			ostream << iter << std::endl;
		}

		ostream << std::flush;

		ostream.close();

	}

}



// -----------------------------------------------



int main(int argc, char* argv[])
{

	// the parameter range
	capd::vectalg::Vector<interval,2> paramRange;
	paramRange(1) = interval(0.01,0.0101);
	paramRange(2) = interval(0.9,0.9001);

	// size of parameter segments to be analyzed
	capd::vectalg::Vector<double,2> paramsize;
	paramsize(1) = 1e-4;
	paramsize(2) = 1e-4;
	
	// initial box size
	double initialBoxSize = 1e-1;

	// initial output files
	std::string successFile = successBase + ".txt";
	std::string errorFile = errorBase + ".txt";

	// we have overriding command line parameters
	if (argc > 1) {
		paramRange(1) = interval(atof(argv[1]), atof(argv[2]));
		paramRange(2) = interval(atof(argv[3]), atof(argv[4]));
		paramsize(1) = atof(argv[5]);
		paramsize(2) = atof(argv[6]);
		initialBoxSize = atof(argv[7]);
		#ifdef FERI_OPENMP
			omp_set_num_threads(atoi(argv[8]));
			if (argc > 9) {
				successFile = successBase + "_" + boost::lexical_cast<std::string>(paramRange(1).leftBound()) + "_" + boost::lexical_cast<std::string>(paramRange(1).rightBound()) + "_" + boost::lexical_cast<std::string>(paramRange(2).leftBound()) + "_" + boost::lexical_cast<std::string>(paramRange(2).rightBound()) + ".txt";
				errorFile = errorBase + "_" + boost::lexical_cast<std::string>(paramRange(1).leftBound()) + "_" + boost::lexical_cast<std::string>(paramRange(1).rightBound()) + "_" + boost::lexical_cast<std::string>(paramRange(2).leftBound()) + "_" + boost::lexical_cast<std::string>(paramRange(2).rightBound()) + ".txt";
			}
		#else
			if (argc > 8) {
				successFile = successBase + "_" + boost::lexical_cast<std::string>(paramRange(1).leftBound()) + "_" + boost::lexical_cast<std::string>(paramRange(1).rightBound()) + "_" + boost::lexical_cast<std::string>(paramRange(2).leftBound()) + "_" + boost::lexical_cast<std::string>(paramRange(2).rightBound()) + ".txt";
				errorFile = errorBase + "_" + boost::lexical_cast<std::string>(paramRange(1).leftBound()) + "_" + boost::lexical_cast<std::string>(paramRange(1).rightBound()) + "_" + boost::lexical_cast<std::string>(paramRange(2).leftBound()) + "_" + boost::lexical_cast<std::string>(paramRange(2).rightBound()) + ".txt";
			}
		#endif
	}



	// printing to succesfile
	std::ofstream ostream(successFile.c_str(), std::ios::app);
	ostream << std::setw(2*oPrecision+7) << "Parameter range" << "      " << std::setw(oPrecision+2) << "Run time" << "      " << std::setw(13) << "Iteration no." << std::setw(28) << "box size" << std::setw(0) << "       enclosure" << std::endl;
	ostream.close();



	// printing to errorfile
	ostream.open(errorFile.c_str(), std::ios::app);
	ostream << std::setw(2*oPrecision+7) << "Parameter range" << "      " << std::setw(oPrecision+2) << "Run time" << "      " << std::setw(13) << "Iteration no." << std::setw(28) << "box size" << std::setw(0) << "       enclosure" << std::endl;
	ostream.close();



	// choice of parameter from range
	long int iter, size, size1, size2;
	
	capd::vectalg::Vector<interval, 2> parameter;
	
	size1 = (long int) (diam(paramRange(1)).rightBound() / paramsize(1)) + 1;
	size2 = (long int) (diam(paramRange(2)).rightBound() / paramsize(2)) + 1;
	size = size1 * size2;

	std::vector< capd::vectalg::Vector<interval, 2> > params;
	params.reserve(size);

	parameter(1) = paramRange(1).left() + interval(0, paramsize(1));
	while (parameter(1).rightBound() < paramRange(1).rightBound()) {

		parameter(2) = paramRange(2).left() + interval(0, paramsize(2));
		while (parameter(2).rightBound() < paramRange(2).rightBound()) {
			if (parameter(2).rightBound() >= (interval(1) - parameter(1)).leftBound()) {
				params.push_back(parameter);
			}
			parameter(2) = interval(parameter(2).rightBound(), parameter(2).rightBound() + paramsize(2));
		}
		parameter(2) = interval(parameter(2).leftBound(), paramRange(2).rightBound());
		if (parameter(2).rightBound() >= (interval(1) - parameter(1)).leftBound()) {
			params.push_back(parameter);
		}

		parameter(1) = interval(parameter(1).rightBound(), parameter(1).rightBound() + paramsize(1));
	}
	
	parameter(1) = interval(parameter(1).leftBound(), paramRange(1).rightBound());
	parameter(2) = paramRange(2).left() + interval(0, paramsize(2));
	while (parameter(2).rightBound() < paramRange(2).rightBound()) {
		if (parameter(2).rightBound() >= (interval(1) - parameter(1)).leftBound()) {
			params.push_back(parameter);
		}
		parameter(2) = interval(parameter(2).rightBound(), parameter(2).rightBound() + paramsize(2));
	}
	parameter(2) = interval(parameter(2).leftBound(), paramRange(2).rightBound());
	if (parameter(2).rightBound() >= (interval(1) - parameter(1)).leftBound()) {
		params.push_back(parameter);
	}

	std::cout.precision(oPrecision);
	std::cout << "Parameter range " << paramRange << " is divided into " << params.size() << " pieces. (paramSize: "<< paramsize << ", initialBoxSize: " << initialBoxSize << ")" << std::endl;



	// checking every parameter chunk
	#ifdef FERI_OPENMP
		#pragma omp parallel shared(params, initialBoxSize)
	#endif
	{
		#ifdef FERI_OPENMP
			#pragma omp for schedule(dynamic)
		#endif
		for(iter = 0; iter < (long int) params.size(); ++iter)
		{
			simulateKW(params[iter](1), params[iter](2), initialBoxSize, successFile, errorFile);
		}
	}

}

// ~ end of file ~
#endif
