#ifndef FERI_RICKER
#define FERI_RICKER
// -----------------------------------------------
// ATTRACTOR ANALYSIS WITH GRAPH REPRESENTATION
//
// Krisztin-Walther Map

// -----------------------------------------------

// MIDPOINT EVALUATION
#define FERI_MIDPOINT

// -----------------------------------------------

#include "main.hpp"
#include "typedefs.hpp"

// -----------------------------------------------

// 2D Ricker Map
template<typename T, int D>
class VectorField : public capd::dynsys::FadMap<T, D>
{
	public:

		typedef typename capd::dynsys::FadMap<T, D> BaseClass;
		typedef typename BaseClass::MatrixType MatrixType;
		typedef typename BaseClass::VectorType VectorType;
		typedef typename BaseClass::TFVector TFVector;


		// constructor
		VectorField ()
		{}

		VectorField (const T &m, const T &ALPHA)
		{
			PARAM_m = m;
			PARAM_alpha = ALPHA;
		}

		// -----------------------------------------------

		// FadBad::vectorfield
		template <typename AVector>
		AVector operator()(const AVector &in) const
		{
			AVector out(2);

			// 1 iterate
			out[0] = in[1];
			out[1] = PARAM_m * in[1] - PARAM_alpha * ( interval(1) - interval(2) / (exp(interval(2) * in[0]) + interval(1)) );
			//out[1] = PARAM_m * in[1] - PARAM_alpha * in[0] * interval(-1,1);
			
			return out;
		}

		// FadBad::derivative
		MatrixType operator[](const VectorType& u) const
		{
			MatrixType der(2, 2);
			computeDerivative(*this, u, der);
			return der;
		}

		// FadBad::returns the dimension
		int getDimension() const
		{
			return D;
		}

		// parameters of the map
		T PARAM_m;
		T PARAM_alpha;

		// evaluate the vectorfield
		void evaluate(const VectorType &in, VectorType &out);

		// constructing the starting invariant set
		void constructInvariantSet(VectorType &V);

		// constructing the local domain of attraction
		void findAttractionDomain(VectorType &attDomain);

		// resetting graphics
		void resetGraphics(const double &ScreenSize);

		// plotting the boxes
		void plotFixedPoint(const double &SCREENSIZE);

	protected:
};

// -----------------------------------------------

// evaluate Function
template<typename T, int D>
void VectorField<T, D>::evaluate(const VectorType &in, VectorType &out)
{
	#ifdef FERI_MIDPOINT
		VectorType in_x;

		in_x = midVector(in);
		out = (*this)(in_x) + (*this)[in] * (in - in_x);
	#else
		out = (*this)(in);
	#endif
}

// constructing the invariant set
template<typename T, int D>
void VectorField<T, D>::constructInvariantSet(VectorType &V)
{
	V(1) = interval(-4,4);
	V(2) = V(1);
}

// LOCAL ATTRACION DOMAIN
template<typename T, int D>
void VectorField<T, D>::findAttractionDomain(VectorType &attDomain)
{
	#ifdef FERI_GRAPHICS_DRAW_EXTRA
		double lt1, lt2, rb1, rb2;
	#endif

	interval epsilon;

	if (PARAM_alpha.leftBound() >= interval(0.98).leftBound()) {
		epsilon = interval(1) / interval(6);
	} else {
		epsilon = sqrt(sqrt(interval(27)/interval(800)))*sqrt(1-sqrt(PARAM_alpha));
	}
	attDomain(1) = interval( -epsilon.leftBound(), epsilon.leftBound() );
	attDomain(2) = attDomain(1);

	#ifdef FERI_GRAPHICS_DRAW_EXTRA

		lt1 = attDomain(1).leftBound();
		lt2 = attDomain(2).leftBound();
		rb1 = attDomain(1).rightBound();
		rb2 = attDomain(2).rightBound();

		rootFrame.box(lt1, lt2, rb1, rb2, YELLOW);
		waitBt();
	#endif
}

// -----------------------------------------------

// resetting graphics
template<typename T, int D>
void VectorField<T, D>::resetGraphics(const double &ScreenSize)
{
	rootFrame.Clear();

	rootFrame.line(- ScreenSize / 2, - ScreenSize / 2, ScreenSize / 2, ScreenSize / 2);
}

// plotting the boxes
template<typename T, int D>
void VectorField<T, D>::plotFixedPoint(const double &SCREENSIZE)
{
	double lt1, lt2, rb1, rb2;
	interval fpoint;


	fpoint = 0 + interval(-1,1) * SCREENSIZE / 100.0;

	lt1 = fpoint.leftBound();
	rb1 = fpoint.rightBound();

	lt2 = lt1;
	rb2 = rb1;

	rootFrame.boxFill(lt1, lt2, rb1, rb2, RED);
	waitBt();
}

// ~ end of file ~
#endif
